#%%
from dcctk.corpusReader import PlainTextReader
from dcctk.concordancer import Concordancer
# from dcctk.corpus import IndexedCorpus
# from dcctk.dispersion import Dispersion
# from dcctk.compoAnalysis import CompoAnalysis
# from dcctk.compoConcordancer import CompoConcordancer
# from dcctk.embeddings import AnchiBert

# c = Dispersion(PlainTextReader("data/").corpus)

# c = TextBasedCorpus(PlainTextReader("data/").corpus)
c = Concordancer(PlainTextReader("data/").corpus)
#%%
cql = """
"夫" "妻"
"""
left_collo = c.collocates(cql, left=1, right=0, subcorp_idx=0, sort_by="Gsq")


# c2 = CompoAnalysis(c)
# c.freq_distr_ngrams(n=2, subcorp_idx=0).most_common(10)
# c2.freq_distr(tp="chr")
# c2.freq_distr(tp="idc")
# c2.freq_distr(tp="rad")
#%%
cql = '''
[ compo="龜" & idc="vert2" & pos="1" ]
'''
cql = '''
[ radical="龜" ] [char="[一-龜]"]
'''
cql = '''
[char="龜"]
'''
cql = '''
[phon="ㄍㄨㄟ" & tone="2"]
'''
cql = '''
[phon=".ㄨㄥ$" & tone="2"]
'''
cql = '''
[phon="^p" & tone="2" & tp="pinyin"] [phon="^ㄆ"]
'''
# cql = '''
# [phon="^pʰ" & tp="ipa"] [phon="^pʰ" & tp="ipa"] [ compo="亻" & idc="horz2" & pos="0" ]
# '''
# cql = '''
# [ compo="龜" & idc="vert2" & pos="1" ]
# '''
cql = '''
[idc="encl"] [idc="encl"]
'''
for i, r in enumerate(c.cql_search(cql)):
    if i == 10: break
    print(r)
# results = list(c.cql_search(cql))
# results[:5]

#%%
c2 = CompoAnalysis(c)

#%%
c2.freq_distr(tp="chr")
#%%
c2.freq_distr(tp="idc")
#%%
c2.freq_distr(tp="rad")
#%%
c2.freq_distr(tp=None, radical="广")
#%%
c2.freq_distr(tp=None, compo="水", idc="vert2", pos=-1)
#%%
c2.productivity(radical="广", subcorp_idx=2, text_idx=1)
#%%
c2.productivity(compo="虫", idc="horz2", pos=0)
#%%
from CompoTree import IDC
for idc in IDC:
    p = c2.productivity(idc=idc.name, subcorp_idx=2)
    print(idc.name, idc.value)
    print(p['productivity'])
    print()
    





#%%
# queries = cqls.parse(
#             cql, default_attr='char', max_quant=3)

# negative_match = set()
# keyword = queries[0][-1]
# chars = keyword['not_match'].get('char', [])
# for idx in c._union_search(chars):
#     negative_match.add(idx)

results = list(c.cql_search(cql, left=10, right=10))
for r in results[:5]: print(r)
#%%
for i, char in enumerate(c._search_keyword(keyword)):
    print(c._get_corp_data(*char))
    if i == 5: break

#%%
from collections import Counter

c.char_dispersion('a', subcorp_idx=None, return_raw=True)
# c.char_dispersion('a', subcorp_idx)
subcorp_idx = None
char = 'a'
v = Counter((i, j) for i, j, _, _ in c.index.get(char, []))
# d = c._get_dispersion_data(v, subcorp_idx)
#stats = { n: func(d) for n, func in c.dispersion_func }

#%%
# c = TextBasedCorpus(PlainTextReader().corpus)
# c.list_files('三國')
# c.get_meta_by_path('03/三國志_蜀書一.txt')
# c.get_text('03/三國志_蜀書一.txt', as_str=True)
texts = c.get_texts('三國志', texts_as_str=False, sents_as_str=True)
texts_str = c.get_texts('三國志', texts_as_str=True, sents_as_str=True)

#%%
cql = '''
"法" "院"
'''.strip()
results = list(c.cql_search(cql, left=10, right=10))
print(len(results))
x = results[0]
#%%
x.get_kwic()
x.get_timestep()

#%%
# Sort Concord
results_sorted = sorted(results, key=lambda x: x.get_timestep())
# results[0].get_timestep()

#%%
emb = AnchiBert()

#%%
base_sent, base_tk = results[17].get_kwic()
print(base_sent)
print(base_tk)
sem_sort_by_tk = semantic_sort(emb, results[:200], base_sent, base_tk)
sem_sort_by_sent = semantic_sort(emb, results[:200], base_sent)

#%%
import gdown


gdown.download('https://drive.google.com/uc?id=1uMlNJzilEhSigIcfjTjPdYOZL9IQfHNK', output="AnchiBERT.zip")

#%%
gdown.extractall("AnchiBERT.zip")

# %%
import cqls
queries = cqls.parse(cql, default_attr=c._cql_default_attr,max_quant=6)

x = c._search_keyword(queries[0][0])
    # for result in c._kwic(keywords=query, left=5, right=5):
    #     results.append(result)
# %%
import math

query = queries[0]

best_search_loc = (0, None, math.inf)
for i, keyword in enumerate(query):
    results = c._search_keyword(keyword)
    num_of_matched = len(results)
    if num_of_matched == 0: 
        pass
    elif num_of_matched < best_search_loc[-1]:
        best_search_loc = (i, results, num_of_matched)
results = best_search_loc[1]


#%%
from dcctk.UtilsConcord import queryMatchToken

keyword_anchor = {
    'length': len(query),
    'seed_idx': best_search_loc[0]
}
keywords = query

matched_results = []
for idx in results:
    # Get all possible matching keywords from corpus
    candidates = c._get_keywords(keyword_anchor, *idx)
    if len(candidates) != len(keywords): 
        continue
    # Check every token in keywords
    matched_num = 0
    for w_k, w_c in zip(keywords, candidates):
        if queryMatchToken(queryTerm=w_k, corpToken=w_c):
            matched_num += 1
    if matched_num == len(keywords):
        first_keyword_idx = idx[2] - keyword_anchor['seed_idx']
        matched_results.append( [idx[0], idx[1], first_keyword_idx] )
# %%



import re
from typing import Union

REGEX_META = set('[].^$*+{}|()')
ESCAPE = chr(92)
SPECIAL_SET = {r'\d', r'\D', r'\s', r'\S', r'\w', r'\W'}
dft_attr = 'char'

def queryMatchToken(queryTerm: dict, corpToken: str):
    if 'match' in queryTerm:
        positive_matched_tag = 0
        for tag, values in queryTerm.get('match').items():
            if tag != dft_attr: continue
            ###### FIX BUG HERE ###############
            if allValues_match_token(values, tag, corpToken): 
                positive_matched_tag += 1
            #######################
        
        if positive_matched_tag != len(queryTerm.get('match')):
            return False
    
    if 'not_match' in queryTerm:
        negative_matched_tag = 0
        counter = 0
        values =  queryTerm.get('not_match').get(dft_attr, [])
        target_value = corpToken
        for value in values:
            counter += 1
            value, mode = match_mode(value)
            if mode == "literal":
                if value != target_value:
                    negative_matched_tag += 1
            else:
                value = append_regex_anchors(value)
                if (target_value != None) and (not re.search(value, target_value)):
                    negative_matched_tag += 1
        
        if negative_matched_tag != counter:
            return False
    
    return True



def allValues_match_token(values:list, tag:Union[str, int] , target: str) -> bool:
    """Check whether all CQL generated values match a token in corpus

    Parameters
    ----------
    values : list
        A list of attribute values to check
    tag : Union[str, int]
        Attribute of the token
    target : str
        A token object retrieved from the corpus

    Returns
    -------
    bool
        Match or not
    """
    matched_num = 0

    for value in values:
        value, mode = match_mode(value)
        if mode == "regex":
            value = append_regex_anchors(value)
            if re.search(value, target):
                matched_num += 1
        if mode == "literal":
            if value == target:
                matched_num += 1
    
    return matched_num == len(values)


def append_regex_anchors(x: str):
    x = "(" + x + ")"
    if not x.startswith('^'):
        x = '^' + x
    if not x.endswith('$'):
        x += '$'
    return x

def match_mode(x: str):
    if is_regex(x):
        try:
            re.compile(x)
            is_valid = True
        except re.error:
            is_valid = False
        if is_valid:
            return x, "regex"
    return x.replace(ESCAPE, ''), "literal"


def is_regex(x: str) -> bool:
    prev = ''
    for this in x:
        if prev + this in SPECIAL_SET:
            return True
        if this in REGEX_META:
            if prev != ESCAPE:
                return True
        
        prev = this
    return False



import pandas as pd
from copy import deepcopy


class SenseAnalysis:
    """Sense distribution based on sense clustering of concordance lines
    """

    sense_distribution = None
    sense_distribution_raw = None
    ts_label_map = []
    """Timestep meta info
    """

    def __init__(self, CS):
        """Initialize sense analysis

        Parameters
        ----------
        CS : ConcordSimil
            An instance of :class:`concordSimil.ConcordSimil` where 
            hierarchical clustering has been performed on.
        """
        self._embeddings = CS.embeddings
        self._embeddings_plotting = CS.embeddings_plotting
        self.df = CS.hover_df

        if 'cluster' not in self.df:
            raise Exception(f'Clustering not performed yet, please run `.hierarchical_clustering()` in {CS} before passing it in.')

        # Sense distribution
        self._sense_distribution_across_time()
    
    
    def plot_sense_distribution(self, timelabel:str=None, **kwargs):
        """Plot distribution of sense clusters across time

        Parameters
        ----------
        timelabel : str, optional
            Key specifying the label to use on the x-axis. Avaliable labels
            are found in :attr:`SenseAnalysis.ts_label_map`. By default None.
        """
        ax = self.sense_distribution.plot.bar(**kwargs)
        if timelabel is None:
            ax.set_xticklabels(list(range(len(self.ts_label_map))), rotation=0)
        else:
            ax.set_xticklabels([l[timelabel] for l in self.ts_label_map], rotation=0)


    def plot_sense_timeseries(self, clusters:list=None, timelabel:str=None, raw_count=True, **kwargs):
        d = deepcopy(self.sense_distribution_raw)
        if not raw_count:
            total_freq = d.apply(lambda x: x.sum(), axis=0)
            d = d / total_freq

        # Subset
        if clusters is not None:
            d = d[ [int(x) for x in clusters] ]
        if timelabel is not None:
            d['label'] = [l[timelabel] for l in self.ts_label_map]
        else:
            d['label'] = [ str(i) for i in range(len(self.ts_label_map)) ]

        ax = d.plot.line(x='label', **kwargs)  # cmap=plt.cm.get_cmap('Reds', d.shape[1])
        ax = ax.set_xlabel('')


    def _sense_distribution_across_time(self):
        # Get frequency distributions
        distr = {}  
        num_of_ts = len(set(self.df.timestep))
        self.ts_label_map = [0] * num_of_ts
        time_meta_labels = self.df.columns[self.df.columns.to_series().str.contains('m.time')]
        for idx, row in self.df.iterrows():
            c = row['cluster']
            ts = int(row['timestep'])
            
            # Get distribution
            if c not in distr:
                distr[c] = [0] * num_of_ts
            distr[c][ts] += 1

            # Get timestep info
            if isinstance(self.ts_label_map[ts], int):
                self.ts_label_map[ts] = dict(row[time_meta_labels])

        d = pd.DataFrame(distr)
        d = d.reindex(sorted(d.columns), axis=1)
        self.sense_distribution_raw = deepcopy(d)
        total_freq = d.apply(lambda x: x.sum(), axis=1)
        for index, row in d.iterrows(): 
            d.loc[index] = row / total_freq[index]
        self.sense_distribution = d



from typing import Sequence
from CompoTree import ComponentTree, Radicals, CharLexicon, IDC, CTFounds
from hanziPhon import Moe, GuangYun
from .UtilsConcord import match_mode

ctree = ComponentTree.load()
radicals = Radicals.load()
match_cache = dict()
radical_map = None
idc_rev_map = None
phon_map = None
idc_val_nm = { x.value:x.name for x in IDC }
idc_names = set(idc_val_nm.values())


def char_match_compo(char:str, tk:dict, lexicon:CharLexicon, hash):
    key = (char, str(tk), hash)
    if key in match_cache:
        return match_cache[key]
    match_cache[key] = False
    if char in find_compo(tk, lexicon, hash):
        match_cache[key] = True
    return match_cache[key]
    

def find_compo(tk:dict, lexicon:CharLexicon, hash):
    """Search hanzi by sub-character features (component, radical, sound) 
    """
    if 'phon' in tk['match'] or 'sys' in tk['match']:
        return phonetic_search(tk, lexicon)
    elif 'radical' in tk['match']:
        return radical_search(tk, lexicon)
    elif 'compo' in tk['match']:
        return component_search(tk, lexicon)
    elif 'idc' in tk['match']:
        return idc_search(tk, lexicon)
    else:
        raise Exception('Invalid CQL attributes.')
    

# Sub-character search functions
def phonetic_search(tk, lexicon):
    """Search hanzi with phonetic representations

        .. code-block:: python

            {
                'match': {
                    'phon': ['ㄅㄨ'],
                    'tone': ['1'],
                    'tp': ['pinyin'],  # bpm, pinyin, ipa
                    'sys': ['moe']
                },
                'not_match': {}
            }
    """
    if tk['match'].get('sys')[0] == '廣韻':
        build_phon_map(lexicon, moe=False, 廣韻=True)
        params = { k:v[0] for k, v in tk['match'].items() if k != 'sys' }
        return phon_map['廣韻'].find(return_raw=False, **params)
    else:
        build_phon_map(lexicon, moe=True, 廣韻=False)
        sp = {
            'phon': '',
            'tone': None,
            'tp': "bpm",
        }
        for k, v in tk['match'].items():
            if k in sp: sp[k] = v[0]
        sys = phon_map['moe']
        phon, mode = match_mode(sp['phon'])
        exact = mode == 'literal'
        return sys.find(repr=phon, tone=sp['tone'], tp=sp['tp'], exact=exact)


def radical_search(tk, lexicon):
    """Search hanzi with radical

    .. code-block:: python
        
        {
            'match': {
                'radical': ['人']
            },
            'not_match': {}
        }
    """
    build_radical_map(lexicon)
    rad = tk['match']['radical'][0]
    return radical_map.get(rad, set())


def component_search(tk, lexicon):
    """Search hanzi with character component

    .. code-block:: python
        
        {
            'match': {
                'compo': ['忄'],
                'idc': ['horz2'],
                'pos': ['0'],
                'max_depth': ['1']
            },
            'not_match': {}
        }
    """
    sp = {
        "compo": '',
        "max_depth": 1,
        'idc': None,  # IDC['horz2'].value,
        'pos': -1,
    }
    for k, v in tk['match'].items():
        if k in sp:
            v = v[0]
            if k in ['max_depth', 'pos']: v = int(v)
            if k == 'idc': v = IDC[v].value
            sp[k] = v
    
    bottom_hits = ctree.find(sp['compo'], max_depth=sp['max_depth'], bmp_only=True)
    if sp['idc'] is None:
        return set( x[0] for x in CTFounds(bottom_hits)\
            .filter_with_lexicon(lexicon)\
            .tolist() )

    return set( x[0] for x in CTFounds(bottom_hits)\
        .filter(idc=sp['idc'], pos=sp['pos'])\
        .filter_with_lexicon(lexicon)\
        .tolist() )


def idc_search(tk, lexicon):
    """Search hanzi with character shape

    .. code-block:: python
        
        {
            'match': {
                'idc': ['horz2'],
            },
            'not_match': {}
        }
    """
    global idc_rev_map
    build_idc_rev_map(lexicon)
    idc = tk['match']['idc'][0]
    if idc not in idc_names: 
        raise Exception(f"Invalid IDC value `{idc}`!", 
                        f"IDC must be one of {', '.join(idc_names)}")
    return idc_rev_map.get(idc, set())


# Helper functions
def get_radicals(lexicon:CharLexicon):
    build_radical_map(lexicon)
    return set(radical_map.keys())


def build_radical_map(lexicon:CharLexicon):
    global radical_map
    if radical_map is None:
        print('Building index for character radicals...')
        radical_map = {}
        for char in lexicon.lexicon:
            rad = radicals.query(char)[0]
            radical_map.setdefault(rad, set()).add(char)


def build_idc_rev_map(lexicon:CharLexicon):
    global idc_rev_map, idc_val_nm
    if idc_rev_map is None:
        print("Building index for character IDCs...")
        idc_rev_map = {}
        for ch in lexicon.lexicon:
            idc = ctree.ids_map.get(ch, [None])[0]
            if idc is None: continue
            idc = idc_val_nm.get(idc.idc, '')
            idc_rev_map.setdefault(idc, set()).add(ch)


def build_phon_map(lexicon:CharLexicon, moe=False, 廣韻=True):
    global phon_map
    if phon_map is None:
        phon_map = {}
    if moe and ('moe' not in phon_map):
        phon_map['moe'] = Moe(lexicon=lexicon.lexicon)
    if 廣韻 and ('廣韻' not in phon_map):
        phon_map['廣韻'] = GuangYun(lexicon=lexicon.lexicon)


def load_lexicon(lexicon: Sequence):
        lexicon = set(lexicon)
        return CharLexicon(lexicon, [], [])



#%%
from itertools import chain
from collections import Counter
from CompoTree import ComponentTree, Radicals, IDC
from CompoTree import CTFounds

ctree = ComponentTree.load()
radicals = Radicals.load()
idc_names = { x.name for x in IDC }

class CompoAnalysis:

    def __init__(self, IndexedCorpus):
        """[summary]

        Parameters
        ----------
        IndexedCorpus : IndexedCorpus
            A corpus with an index of character positions. Accepts 
            :class:`dcctk.corpus.IndexedCorpus` or its super classes as inputs.
        """
        self.corpus = IndexedCorpus.corpus
        self.index = IndexedCorpus.index
        self.cc_map = {}
        self.rad_map = {}
        self.idc_map = {}
        self.fq_distr_cache = {}
        self.corp_fq_info = {}
        self._build_cc_map()
        self._build_rad_map()
        self._build_idc_map()


    def productivity(self, radical=None, compo=None, idc=None, pos=-1, 
                     subcorp_idx:int=None, text_idx:int=None):

        chr_fq = self.freq_distr(subcorp_idx, text_idx, "chr")
        if radical:
            chars = self.rad_map.get(radical, set())
        elif compo:
            chars = self._component_search(compo, idc, pos)
        elif idc:
            chars = self._idc_search(idc)
        else:
            raise Exception("One of `radical`, `compo`, or `idc` must be given")
        chars.intersection_update(chr_fq.keys())

        V1C, NC = 0, 0
        for ch in chars:
            fq = chr_fq[ch]
            NC += fq
            if fq == 1: V1C += 1
        
        # Static info
        k = (subcorp_idx, text_idx)
        if k not in self.corp_fq_info:
            self.corp_fq_info[k] = {
                'N': sum(chr_fq.values()),
                'V1': sum(f for f in chr_fq.values() if f == 1)
            }

        V1 = self.corp_fq_info[k]['V1']
        return {
            'productivity': {
                'realized': len(chars),
                'expanding': V1C / V1,
                'potential': V1C / NC if V1C != 0 else 0,
            },
            'N': self.corp_fq_info[k]['N'],
            'V1': V1,
            'V1C': V1C,
            'NC': NC
        }

    def _component_search(self, compo:str, idc=None, pos:int=-1):
        bottom_hits = ctree.find(compo, max_depth=1, bmp_only=True)
        if idc is None:
            return set( x[0] for x in CTFounds(bottom_hits)\
                .tolist() )
        return set( x[0] for x in CTFounds(bottom_hits)\
            .filter(idc=IDC[idc].value, pos=pos)\
            .tolist() )

    
    def _idc_search(self, idc:str="vert2"):
        global idc_names
        if idc not in idc_names: 
            raise Exception(f"Invalid IDC value `{idc}`!", 
                            f"IDC must be one of {', '.join(idc_names)}")
        return self.idc_map.get(idc, set())



    def freq_distr(self, subcorp_idx=None, text_idx=None, tp="idc", 
                   compo=None, idc=None, pos=-1, radical=None):
        """Frequency distribution of character (component)

        Parameters
        ----------
        subcorp_idx : int, optional
            Index for subcorpus, by default None, which uses the whole corpus.
        text_idx : int, optional
            Index for text in a subcorpus, by default None, which uses the 
            whole subcorpus.
        tp : str, optional
            One of :code:`chr` (Character), :code:`idc` 
            (Ideographic Description Characters), :code:`rad` (Radical), 
            and None, by default :code:`idc`

        Returns
        -------
        Counter
            A freqeuncy distribution.
        """
        # Use cache
        k = (subcorp_idx, text_idx, tp, compo, idc, pos, radical)
        if k in self.fq_distr_cache: return self.fq_distr_cache[k]

        # Character frequency distribution
        if tp == 'chr' or tp == 'char':
            self.fq_distr_cache[k] = self._freq_distr_chr(subcorp_idx, text_idx)
            return self.fq_distr_cache[k]
        
        # Character component frequency distribution
        fq_ch = self._freq_distr_chr(subcorp_idx, text_idx)
        if tp == "idc" or tp == "rad":
            fq_compo = Counter()
            for ch, fq in fq_ch.items():
                k = "noChrData"
                if ch in self.cc_map:
                    k = self.cc_map[ch].get(tp, "noCompoData")
                fq_compo.update({k: fq})
            self.fq_distr_cache[k] = fq_compo
            return fq_compo
        
        # Token distribution within a radical/component
        if radical:
            chars = self.rad_map.get(radical, set())
            chars.intersection_update(fq_ch.keys())
            return Counter({ch:fq for ch, fq in fq_ch.items() if ch in chars})
        if compo:
            chars = self._component_search(compo, idc, pos)
            chars.intersection_update(fq_ch.keys())
            return Counter({ch:fq for ch, fq in fq_ch.items() if ch in chars})


    def _freq_distr_chr(self, subcorp_idx:int=None, text_idx:int=None):
        if isinstance(text_idx, int) and isinstance(subcorp_idx, int):
            corp = self.corpus[subcorp_idx]['text'][text_idx]['c']
            return Counter(chain.from_iterable(corp))
        if isinstance(subcorp_idx, int):
            corp = (c for t in self.corpus[subcorp_idx]['text'] for c in t['c'])
            return Counter(chain.from_iterable(corp))
        
        corp = (c for sc in self.corpus for t in sc['text'] for c in t['c'])
        return Counter(chain.from_iterable(corp))


    def _build_cc_map(self):
        for ch in self.index:
            idc = ctree.ids_map.get(ch, [None])[0]
            rad = radicals.query(ch)[0]
            dph = None
            if idc is not None:
                idc = idc.idc
                dph = ctree.query(ch, use_flag="shortest", max_depth=-1)[0]
                if not isinstance(dph, str):
                    dph = dph.depth()
                else:
                    dph = 0
            if idc is None and rad == '': continue
            self.cc_map[ch] = {
                'idc': idc,
                'rad': rad,
                'dph': dph
            }
    
    def _build_rad_map(self):
        for ch in self.cc_map:
            r = self.cc_map[ch]['rad']
            self.rad_map.setdefault(r, set()).add(ch)

    def _build_idc_map(self):
        idc_val_nm = { x.value: x.name for x in IDC }
        for ch in self.index.keys():
            idc = ctree.ids_map.get(ch, [None])[0]
            if idc is None: continue
            idc = idc_val_nm.get(idc.idc)
            if idc:
                self.idc_map.setdefault(idc, set()).add(ch)



import pandas as pd
from itertools import chain
from collections import Counter
from typing import Union, Sequence
from .corpus import IndexedCorpus
from .dispersionStats import Range, DP, DPnorm, KLdivergence, JuillandD, RosengrenS


class Dispersion(IndexedCorpus):
    
    def __init__(self, corpus) -> None:
        IndexedCorpus.__init__(self, corpus)

        # Corpus sizes
        self.num_of_subcorp = len(self.corpus)
        self.corp_size = 0
        self.subcorp_size = [ 0 ] * self.num_of_subcorp
        self.text_size = {}
        self.num_of_text = 0
        self.num_of_text_in_each_subcorp = [ 0 ] * self.num_of_subcorp
        self._compute_sizes()
        # Free memory
        del self.corpus

        # Functions for measures of dispersion
        self.dispersion_func = [ 
            Range, DP, DPnorm, KLdivergence, JuillandD, RosengrenS
        ]


    def pattern_dispersion(self, data:Union[pd.DataFrame, Sequence], subcorp_idx:int=None, return_raw=False):
        """Compute dispersion measures of a complex pattern

        Parameters
        ----------
        data : Union[pd.DataFrame, Sequence]
            Data returned by :meth:`concordancer.Concordancer.cql_search` 
            (i.e., a sequence/generator of :class:`concordancer.ConcordLine`) 
            or :class:`pandas.DataFrame` with a column :code:`m.id` specifying
            the path to a text file (i.e., a dataframe recording the clustering 
            result of :meth:`concordSimil.ConcordSimil.hierarchical_clustering`).
        subcorp_idx : int, optional
            The index of the subcorpus used to derive the data, 
            by default None, which uses the full corpus to derive 
            the data.
        return_raw : bool, optional
            Whether to return the data used for computing 
            dispersion statisitcs, by default False

        Returns
        -------
        dict or tuple
            A dictionary recording various dispersion measures.
            If :code:`return_raw` is True, an additional dictionary
            recording data used for computing dispersion measures 
            is returned.
        """

        if isinstance(data, pd.DataFrame):
            v = Counter(self.path_index[p] for p in data['m.id'])
        else:
            v = Counter(c.data['position'][:2] for c in data)
        return self._compute_dispersion(v, subcorp_idx, return_raw)


    def char_dispersion(self, char:str, subcorp_idx:int=None, return_raw=False):
        """Compute dispersion measures of a character

        Parameters
        ----------
        char : str
            The target character
        subcorp_idx : int, optional
            The index of the subcorpus used to derive the data, 
            by default None, which uses the full corpus to derive 
            the data.
        return_raw : bool, optional
            Whether to return the data used for computing 
            dispersion statisitcs, by default False

        Returns
        -------
        dict or tuple
            A dictionary recording various dispersion measures.
            If :code:`return_raw` is True, an additional dictionary
            recording data used for computing dispersion measures 
            is returned.
        """
        v = Counter((i, j) for i, j, _, _ in self.index.get(char, []))
        if len(v) == 0:
            print(f'No instances of `{char}` found in the corpus!')
        return self._compute_dispersion(v, subcorp_idx, return_raw)


    def _compute_dispersion(self, v, subcorp_idx, return_raw):
        d = self._get_dispersion_data(v, subcorp_idx)
        stats = { func.__name__: func(d) for func in self.dispersion_func }
        if return_raw:
            return stats, d
        else:
            return stats
    

    def _get_dispersion_data(self, v:Union[Counter, dict], subcorp_idx:int=None):
        """Get data from calculating dispersion statistics

        Parameters
        ----------
        v : Union[Counter, dict]
            A counter with keys in the form of :code:`(<subcorp_idx>, <text_idx>)`
            specifying a text and values raw frequencies of a particular pattern 
            in this text.
        subcorp_idx : int, optional
            The index of the subcorpus used to derive the data, 
            by default None, which uses the full corpus to derive 
            the data.

        Returns
        -------
        dict
            A dictionary of parameters for calculating dispersion stats

        Notes
        -----
        See page 102 in [1] for details of the definitions of the returned data

        References
        ----------
        [1] Gries (2020). Analyzing dispersion. In M. Paquot & S. Th. Gries (Eds.),
        A practical handbook of corpus linguistics (pp. 99–118). Cham: Springer 
        International Publishing. https://doi.org/10.1007/978-3-030-46216-1_5
        """

        if isinstance(subcorp_idx, int):
            corp_size = self.subcorp_size[subcorp_idx]
            size = self.num_of_text_in_each_subcorp[subcorp_idx]
            V = [ 0 ] * size
            P = V.copy()
            S = V.copy()
            for j in self.text_size[subcorp_idx]:
                Vj = v.get((subcorp_idx, j), 0)
                Pj = Vj / self.text_size[subcorp_idx][j]['r']
                Sj = self.text_size[subcorp_idx][j]['s']
                V[j], P[j], S[j] = Vj, Pj, Sj
        else:
            corp_size = self.corp_size
            size = self.num_of_text
            V = [ 0 ] * size
            P = V.copy()
            S = V.copy()
            c = 0
            for i in self.text_size:
                for j in self.text_size[i]:
                    Vj = v.get((i, j), 0)
                    Pj = Vj / self.text_size[i][j]['r']
                    Sj = self.text_size[i][j]['t']
                    V[c], P[c], S[c] = Vj, Pj, Sj
                    c += 1
        return {
                'n': size,
                'v': V,
                'p': P,
                's': S,
                'f': sum(V),
                'corpus_size': corp_size 
        }


    def _compute_sizes(self):
        for i in range(self.num_of_subcorp):
            subcorp_size = 0
            if i not in self.text_size:
                self.text_size[i] = {}
            for j, text in enumerate(self.corpus[i]['text']):
                self.text_size[i][j] = {
                    'r': sum(len(sent) for sent in text.get('c', [])),
                    's': 0.0,
                    't': 0.0,
                }
                subcorp_size += self.text_size[i][j]['r']
            self.subcorp_size[i] = subcorp_size
        self.corp_size = sum(self.subcorp_size)
        self.num_of_text = len(list(chain.from_iterable(self.text_size.values())))

        # Get relative freq
        for i in self.text_size:
            for j in self.text_size[i]:
                self.text_size[i][j]['s'] = self.text_size[i][j]['r'] / self.subcorp_size[i]
                self.text_size[i][j]['t'] = self.text_size[i][j]['r'] / self.corp_size

                self.num_of_text_in_each_subcorp[i] += 1



#%%
import cqls
from itertools import chain
from CompoTree import IDC
from collections import Counter

from .concordancerBase import ConcordancerBase, ConcordLine
from .subCharQuery import find_compo, load_lexicon, char_match_compo, get_radicals
from .UtilsConcord import queryMatchToken
from .UtilsSubchar import all_plain_cql, has_cql_match_type, is_subchar
from .UtilsStats import additive_smooth


class Concordancer(ConcordancerBase):
    
    lexicon = None

    def cql_search(self, cql: str, left=5, right=5):
        queries = cqls.parse(cql, default_attr=self._cql_default_attr, \
            max_quant=self._cql_max_quantity)
        
        # Cond0: all plain cql search
        if sum(all_plain_cql(q) for q in queries) == len(queries):
            for res in super(Concordancer, self).cql_search(cql, left, right):
                yield res
            return
        
        if self.lexicon is None:
            self.lexicon = load_lexicon(self.index.keys())

        for query in queries:
            # Cond1: has plain char match (plain char as anchor)
            if has_cql_match_type(query, "literal"):

                subchar_idx = [i for i, tk in enumerate(query) if is_subchar(tk)]
                query_pl = [q if i not in subchar_idx else {} \
                    for i, q in enumerate(query)]
                
                for result in self._kwic(keywords=query_pl, left=left, right=right):
                    candi = result.data['keyword']
                    matched_num = sum(1 for i in subchar_idx \
                        if char_match_compo(candi[i], query[i], self.lexicon, self.__hash__()) )
                    if matched_num == len(subchar_idx):
                        yield result
            
            # Cond2: no plain char match (compo as anchor)
            else:
                subchar_idx = [i for i, tk in enumerate(query) if is_subchar(tk)]
                tk0 = query[subchar_idx[0]]

                matched_chars = find_compo(tk0, self.lexicon, self.__hash__())

                len_query = len(query)
                keyword_anchor = {
                    'length': len_query,
                    'seed_idx': subchar_idx[0]
                }

                for idx in chain.from_iterable(self.index[c] for c in matched_chars):
                    candidates = self._get_keywords(keyword_anchor, *idx)
                    if len(candidates) != len_query: continue
                    # Check every token in keywords
                    matched_num = 0
                    for w_k, w_c in zip(query, candidates):
                        if is_subchar(w_k):
                            if char_match_compo(w_c, w_k, self.lexicon, self.__hash__()):
                                matched_num += 1
                        else: 
                            if queryMatchToken(queryTerm=w_k, corpToken=w_c):
                                matched_num += 1
                    if matched_num == len_query:
                        subcorp_idx, text_idx, sent_idx, tk_idx = \
                        idx[0], idx[1], idx[2], idx[3] - keyword_anchor['seed_idx']
                        cc = self._kwic_single(subcorp_idx, text_idx, sent_idx, tk_idx, \
                            tk_len=len_query, left=left, right=right, keywords=query)
                        yield ConcordLine(cc)
    

    @property
    def chr_idcs(self):
        return { x.name: x.value for x in IDC }

    @property
    def chr_radicals(self):
        if self.lexicon is None:
            self.lexicon = load_lexicon(self.index.keys())
        return get_radicals(self.lexicon)


    @property
    def cql_attrs(self):
        return {
            "Default": ['char'],
            "CharComponent": ['compo', 'max_depth', 'idc', 'pos'],
            "CharRadical": ['radical'],
            "CharPhonetic": {
                "moe": ['phon', 'tone', 'tp', 'sys="moe"'],
                "廣韻": [
                    '攝', '聲調', '韻母', '聲母', '開合', 
                    '等第', '反切', '拼音', 'IPA', 'sys="廣韻"'
                ]
            }
        }


    def collocates(self, node_cql:str, left=1, right=1, subcorp_idx=None, sort_by="Gsq", alpha=0.1, chinese_only=True):
        """
                     node       ~node
        collocate    O11         O12    R1 (char index len)
        ~collocate   O21         O22    R2
                      C1          C2   CorpSize
                   (Concord num)
        """
        o11 = Counter()
        C1 = 0
        for m in self.cql_search(node_cql, left, right):
            m = m.data
            if isinstance(subcorp_idx, int):
                if m['position'][0] != subcorp_idx: continue
            for w in m["left"] + m["right"]:
                o11.update({w: 1})
            C1 += 1
        
        collo_margin = {}
        for w in o11:
            if chinese_only:
                if self.pat_ch_chr.search(w) is None: continue
            if isinstance(subcorp_idx, int):
                collo_margin[w] = len([1 for i in self.index[w] if i[0] == subcorp_idx])
            else:
                collo_margin[w] = len(self.index[w])
        
        output = []
        N = self._get_corp_size(subcorp_idx)
        for w, R1 in collo_margin.items():
            R2 = N - R1
            O11 = o11.get(w, 0)
            O21 = C1 - O11
            O12 = R1 - O11
            O22 = R2 - O21
            O11, O12, O21, O22, E11, E12, E21, E22 = additive_smooth(O11, O12, O21, O22, alpha=alpha)
            stats = { 
                func.__name__: func(O11, O12, O21, O22, E11, E12, E21, E22)\
                    for func in self.association_measures
            }
            stats['RawCount'] = o11.get(w, 0)
            output.append((w, stats))
        
        return sorted(output, reverse=True, key=lambda x: x[1][sort_by])


    corp_size = None
    def _get_corp_size(self, subcorp_idx=None):
        if self.corp_size is None:
            self.corp_size = {}
            for i in range(len(self.corpus)):
                corp = (c for t in self.corpus[i]['text'] for c in t['c'])
                self.corp_size[i] = len(list(chain.from_iterable(corp)))
        if isinstance(subcorp_idx, int):
            return self.corp_size[subcorp_idx]
        return sum(self.corp_size.values())



import re
import math
import json
import cqls
from typing import Callable
from itertools import chain
from collections import Counter
from .UtilsConcord import queryMatchToken, match_mode
from .corpus import IndexedCorpus


class ConcordancerBase(IndexedCorpus):
    """Retrive concordance lines from the corpus with Corpus Query Language
    """

    _cql_default_attr = "char"
    _cql_max_quantity = 5
    all_idx_cache = None

    def cql_search(self, cql: str, left=5, right=5):
        """Search the corpus with Corpus Query Language

        Parameters
        ----------
        cql : str
            A CQL query
        left : int, optional
            Left context size, by default 5
        right : int, optional
            Right context size, by default 5

        Yields
        -------
        dict
            A dictionary with the structure:

            .. code-block:: python

                {
                    'left': [<tk>, <tk>, ...],
                    'keyword': [<tk>, <tk>, ...],
                    'right': [<tk>, <tk>, ...],
                    'position': ( <int>, <int>, <int>, <int> ),
                    'captureGroups': {
                        'verb': [<tk>],
                        'noun': [<tk>]}
                }

            where ``<tk>`` is a token (char), represented as a string,
        """
        queries = cqls.parse(
            cql, default_attr=self._cql_default_attr, max_quant=self._cql_max_quantity)

        for query in queries:
            for result in self._kwic(keywords=query, left=left, right=right):
                yield result

    def _kwic(self, keywords: list, left=5, right=5):
        # Get concordance from corpus
        search_results = self._search_keywords(keywords)
        if search_results is None:
            return []
        for subcorp_idx, text_idx, sent_idx, tk_idx in search_results:
            cc = self._kwic_single(subcorp_idx, text_idx, sent_idx, tk_idx, tk_len=len(
                keywords), left=left, right=right, keywords=keywords)
            yield ConcordLine(cc)

    def _kwic_single(self, subcorp_idx, text_idx, sent_idx, tk_idx, tk_len=1, left=5, right=5, keywords: list = None):
        # Flatten doc sentences to a list of tokens
        doc = self._get_corp_data(subcorp_idx, text_idx)
        text, keyword_idx = flatten_doc_to_sent(doc['c'])

        tk_start_idx = keyword_idx(sent_idx, tk_idx)
        tk_end_idx = tk_start_idx + tk_len
        start_idx = max(tk_start_idx - left, 0)
        end_idx = min(tk_end_idx + right, len(text))

        # Get CQL labeled token positions
        captureGroups = {}
        for i, keyword in enumerate(keywords):
            if '__label__' in keyword:
                for lab in keyword.get('__label__'):
                    if lab not in captureGroups:
                        captureGroups[lab] = {
                            's': '',
                            'i': [i, i]
                        }
                    tk = self._get_corp_data(
                        subcorp_idx, text_idx, sent_idx, i + tk_idx)
                    captureGroups[lab]['s'] += tk
                    captureGroups[lab]['i'][-1] = i

        return {
            "left": text[start_idx:tk_start_idx],
            "keyword": text[tk_start_idx:tk_end_idx],
            "right": text[tk_end_idx:end_idx],
            "position": (subcorp_idx, text_idx, sent_idx, tk_idx),
            "meta": {
                'id': self.corpus[subcorp_idx]['text'][text_idx]['id'],
                'time': self.get_meta(subcorp_idx, include_id=False),
                'text': self.get_meta(subcorp_idx, text_idx, include_id=False)
            },
            "captureGroups": captureGroups
        }

    def _search_keywords(self, keywords: list):
        #########################################################
        # Find keywords with the least number of matching results
        #########################################################
        best_search_loc = (0, None, math.inf)
        for i, keyword in enumerate(keywords):
            # Skip regex searches
            has_regex = False
            chars = keyword.get('match', {}).get(self._cql_default_attr, []) + \
                keyword.get('not_match', {}).get(self._cql_default_attr, [])
            for char in chars:
                if match_mode(char)[1] == 'regex':
                    has_regex = True
                    break
            if has_regex or len(chars) == 0: continue

            results = self._search_keyword(keyword)
            num_of_matched = len(results)
            if num_of_matched == 0:
                return None
            elif num_of_matched < best_search_loc[-1]:
                best_search_loc = (i, results, num_of_matched)
        results = best_search_loc[1]
        if results is None:
            results = self._search_keyword(keyword)

        #######################################
        # Check other tokens around search seed
        #######################################
        keyword_anchor = {
            'length': len(keywords),
            'seed_idx': best_search_loc[0]
        }

        # Check all possible matching keywords
        matched_results = []
        for idx in results:
            # Get all possible matching keywords from corpus
            candidates = self._get_keywords(keyword_anchor, *idx)
            if len(candidates) != len(keywords):
                continue
            # Check every token in keywords
            matched_num = 0
            for w_k, w_c in zip(keywords, candidates):
                if queryMatchToken(queryTerm=w_k, corpToken=w_c):
                    matched_num += 1
            if matched_num == len(keywords):
                first_keyword_idx = idx[3] - keyword_anchor['seed_idx']
                matched_results.append(
                    [idx[0], idx[1], idx[2], first_keyword_idx])

        return matched_results

    def _search_keyword(self, keyword: dict):
        """Global search of a keyword to find candidates of correct kwic instances

        Parameters
        ----------
        keyword : dict
            A dictionary specifying the matching conditions of
            the keyword:

            .. code-block:: python

                {
                    'match': {
                        'word': ['打'],
                        'pos': ['V.*']
                    }, 
                    'not_match': {
                        'pos': ['VH.*']
                    },
                    '__label__': ['l1']  #labels to attached to search results  
                }

        Returns
        -------
        list
            A list of matching indicies
        """
        positive_match = set()
        negative_match = set()

        # Deal with empty token {}
        if ('match' not in keyword) and ('not_match' not in keyword):
            if self.all_idx_cache is None:
                self.all_idx_cache = set(chain.from_iterable(self.index.values()))
            return self.all_idx_cache
        else:
            ########################################
            ##########   POSITIVE MATCH   ##########
            ########################################
            matching_idicies = Counter()
            chars = keyword['match'].get(self._cql_default_attr, [])
            for idx in self._intersect_search(chars):
                matching_idicies.update({idx: 1})

            # Get indicies that matched all given tags
            for idx, count in matching_idicies.items():
                if count == len(keyword['match']):
                    positive_match.add(idx)

            # Special case: match is empty
            if len(keyword['match']) == 0:
                if self.all_idx_cache is None:
                    self.all_idx_cache = set(chain.from_iterable(self.index.values()))
                positive_match = self.all_idx_cache

            ########################################
            ##########   NEGATIVE MATCH   ##########
            ########################################
            chars = keyword['not_match'].get(self._cql_default_attr, [])
            for idx in self._union_search(chars):
                negative_match.add(idx)

            ########################################
            #####  POSITIVE - NEGATIVE MATCH  ######
            ########################################
            positive_match.difference_update(negative_match)

        if len(positive_match) == 0:
            print(f"{keyword} not found in corpus")

        return positive_match


    def _union_search(self, values: list):
        """Given candidates values, return from corpus the 
        position of tokens matching any of the values.

        Parameters
        ----------
        values : list
            A list of values to compare with
        """
        matched_indicies = set()

        for value in values:
            value, mode = match_mode(value)
            if mode == "literal":
                matched_indicies.update(self.index.get(value, []))
            else:
                for char in self.index:
                    if re.search(value, char):
                        matched_indicies.update(self.index[char])

        return matched_indicies

    def _intersect_search(self, values: list):
        """Given candidates values, return from corpus the
        position of tokens matching all values.

        Parameters
        ----------
        values : list
            A list of values to compare with
        """
        # Get intersections of all values
        match_count = Counter()
        for value in values:
            value, mode = match_mode(value)
            indices = set()
            if mode == "literal":
                indices = self.index.get(value, set())
            else:
                for char in self.index:
                    if re.search(value, char):
                        indices.update(self.index[char])

            for idx in indices:
                match_count.update({idx: 1})

        # Filter idicies that match all values given
        intersect_match = set()
        for idx, count in match_count.items():
            if count == len(values):
                intersect_match.add(idx)

        return intersect_match

    def _get_keywords(self, search_anchor: dict, subcorp_idx, doc_idx, sent_idx, tk_idx):
        sent = self._get_corp_data(subcorp_idx, doc_idx, sent_idx)
        start_idx = max(0, tk_idx - search_anchor['seed_idx'])
        end_idx = min(start_idx + search_anchor['length'], len(sent))

        return sent[start_idx:end_idx]

    def _get_corp_data(self, subcorp_idx, doc_idx=None, sent_idx=None, tk_idx=None):
        """Get corpus data by position
        """
        if doc_idx is None:
            return self.corpus[subcorp_idx]
        if sent_idx is None:
            return self.corpus[subcorp_idx]['text'][doc_idx]
        if tk_idx is None:
            return self.corpus[subcorp_idx]['text'][doc_idx]['c'][sent_idx]
        return self.corpus[subcorp_idx]['text'][doc_idx]['c'][sent_idx][tk_idx]


class ConcordLine:

    def __init__(self, cc: dict):
        """Initialize an instance of concordance line

        Parameters
        ----------
        cc : dict
            A dictionary returned by
            :meth:`concordancer.Concordancer._kwic_single`.
            It has the following stucture:

            .. code-block:: python

                {
                    'left': '，又喜',
                    'keyword': '將軍之去，',
                    'right': '計必乘',
                    'position': (2, 55, 5, 208),
                    'meta': {
                        'id': '03/三國志_蜀書七.txt',
                        'time': {
                            'time_range': [221, 589], 
                            'label': '魏晉南北', 
                            'ord': 3
                        },
                        'text': {
                            'book': '三國志', 'sec': '蜀書七'
                        }
                    },
                    'captureGroups': {
                        'obj': {'s': '去，', 'i': [3, 4]}
                    }
                }
        """
        self.data = cc
        self.meta = obj(**cc.get('meta'))
    

    def __repr__(self) -> str:
        l, k, r = self.data['left'], self.data['keyword'], self.data['right']
        cc = '<Concord ' + l + '{' + k + '}' + r + '>'
        return cc


    def get_kwic(self, return_keyword_idx=True):
        """Get string representation of the concordance line

        Parameters
        ----------
        return_keyword_idx : bool, optional
            Whether to return the index of the keywords
            in the concordance line, by default True

        Returns
        -------
        str or tuple
            If return keyword_idx is True, returns tuple,
            with the second element being the index of the
            keywords (idx_from, idx_to).
        """
        l, k, r = self.data['left'], self.data['keyword'], self.data['right']
        s = l + k + r
        if return_keyword_idx:
            i_fr = len(l)
            i_to = i_fr + len(k) - 1
            return s, (i_fr, i_to)
        return s        


    def get_timestep(self, key:Callable=None):
        """Get time step info of the concordance line

        Parameters
        ----------
        key : Callable, optional
            If specified, applied on self.meta.time to return
            time step data. By default None, which uses 
            subcorp_idx as time step information.

        Returns
        -------
        Int
            The time step that the concordance line belongs to.
        """
        if key is not None:
            return key(self.meta.time)
        return self.data['position'][0]


    def to_json(self, ensure_ascii=False, indent=False):
        return json.dumps(self.data, ensure_ascii=ensure_ascii, indent=indent)



class obj(object):
    def __init__(self, **entries):
        self.__dict__.update(entries)
    
    def __repr__(self):
        s = []
        for k, v in self.__dict__.items():
            s.append(f".{k} = {str(v)}")
        return '\n'.join(s)

##################
# Helper functions
##################
def flatten_doc_to_sent(doc):
    text = ''
    sent_lengths = []
    for sent in doc:
        sent_lengths.append(len(sent))
        text += sent

    def keyword_idx(sent_idx, tk_idx):
        nonlocal sent_lengths
        for i in range(sent_idx):
            tk_idx += sent_lengths[i]
        return tk_idx

    return text, keyword_idx



import json
from pathlib import Path
from .UtilsTextProcess import *


class PlainTextReader:
    """Plain text corpus input handler

    Examples
    --------
    
    .. code-block:: python

        >>> from pprint import pprint
        >>> from gdown import cached_download
        >>> from dcctk.corpusReader import PlainTextReader

        >>> url = 'https://github.com/liao961120/dcctk/raw/main/test/minimal_plaintext_corpus.zip'
        >>> cached_download(url, "minimal_plaintext_corpus.zip", postprocess=gdown.extractall)
        >>> corpus = PlainTextReader("minimal_plaintext_corpus/").corpus
        >>> pprint(corpus)
        
        [{'id': '01',
        'm': {'label': '1st timestep', 'ord': 1, 'time_range': [-1000, -206]},
        'text': [{'c': ['這是第三篇裡的一個句子。', '這是第二個句子。'],
                    'id': '01/text3.txt',
                    'm': {'about': 'Text 3 in 1st timestep'}},
                {'c': ['這是一個句子。', '這是第二個句子。'],
                    'id': '01/text1.txt',
                    'm': {'about': 'Text 1 in 1st timestep'}},
                {'c': ['這是第二篇裡的一個句子。', '這是第二個句子。'],
                    'id': '01/text2.txt',
                    'm': {'about': 'Text 2 in 1st timestep'}}]},
        {'id': '02',
        'm': {'label': '2nd timestep', 'ord': 2, 'time_range': [-205, 220]},
        'text': [{'c': ['這是第三篇裡的一個句子。', '這是第二個句子。'],
                    'id': '02/text3.txt',
                    'm': {'about': 'Text 3 in 2nd timestep'}},
                {'c': ['這是一個句子。', '這是第二個句子。'],
                    'id': '02/text1.txt',
                    'm': {'about': 'Text 1 in 2nd timestep'}},
                {'c': ['這是第二篇裡的一個句子。', '這是第二個句子。'],
                    'id': '02/text2.txt',
                    'm': {'about': 'Text 2 in 2nd timestep'}}]}]
    """

    def __init__(self, dir_path="data/", ts_meta_filename="time.yaml", \
        text_meta_filename="text_meta.yaml", ts_meta_loader=None, 
        text_meta_loader=None, plain_text_reader=read_text_as_sentences):
        """Read in plain text corpus

        Parameters
        ----------
        dir_path : str, optional
            Path to the directory containing the plain text corpus. For the
            directory structure of the plain text corpus, refer to the example
            data in the GitHub `repo`_. By default "data/".

            .. _repo: https://github.com/liao961120/dcctk/tree/main/test/data
        ts_meta_filename : str, optional
            Path to the metadata file specifying the time info of each 
            timestepped subcorpora, by default "time.yaml".
        text_meta_filename : str, optional
            Path to the metadata file specifying info of each corpus text, 
            by default "text_meta.yaml".
        ts_meta_loader : Callable, optional
            Custom function to parse the file specified in 
            :code:`ts_meta_filename`, by default None.
        text_meta_loader : Callable, optional
            Custom function to parse the file specified in 
            :code:`text_meta_filename`, by default None.
        plain_text_reader : Callable, optional
            Function to read a corpus text file as a sequence of sentences, 
            by default :func:`dcctk.UtilsTextProcess.read_text_as_sentences`.
        """
        # Attributes
        self.corp_path = Path(dir_path)
        self.corpus = []
        self.text_meta = {}
        self.timestep_meta = {}
        self.plain_text_reader = plain_text_reader
        self.timestep_meta = self._get_meta(self.corp_path / ts_meta_filename, custom_loader=ts_meta_loader)
        self.text_meta = self._get_meta(self.corp_path / text_meta_filename, custom_loader=text_meta_loader)
        self._read_corpus()


    def _read_corpus(self):
        for dir_ in sorted(self.corp_path.iterdir()):
            if not dir_.is_dir(): continue
            
            # Construct corpus
            meta = {}
            for k, v in self.timestep_meta.get(dir_.stem, {}).items():
                meta[k] = v
            corpus = {
                'id': dir_.stem,
                'm': meta,
                'text': []
            }

            # Read text
            for fp in dir_.glob("*.txt"):
                text_id, text_meta, text_content = self._read_text(fp)
                text = {
                    'id': text_id,
                    'm': text_meta,
                    'c': list(text_content)  # A list of sentences
                }
                corpus["text"].append(text)

            self.corpus.append(corpus)


    def _read_text(self, fp):
        meta = {}
        fp_key = f"{fp.parent.stem}/{fp.name}"
        for k, v in self.text_meta.get(fp_key, {}).items():
            meta[k] = v

        return fp_key, meta, self.plain_text_reader(fp)


    def _get_meta(self, fp, custom_loader=None):
        if not fp.exists(): return {}
        with open(fp, encoding="utf-8") as f:
            if custom_loader:
                return custom_loader()
            elif fp.name.endswith('yaml'):
                import yaml
                return yaml.load(f, Loader=yaml.FullLoader)
            elif fp.name.endswith('json'):
                return json.load(f)



import re
from collections import Counter
from tqdm.auto import trange
from .utils import ngrams
from .UtilsStats import MI, Xsq, Gsq, Dice, DeltaP12, DeltaP21, FisherExact, additive_smooth

class TextBasedCorpus:
    """Corpus object for text-based (text as unit) analysis
    """
    
    association_measures = [
        MI, Xsq, Gsq, Dice, DeltaP21, DeltaP12, FisherExact
    ]

    def __init__(self, corpus):
        self.corpus = corpus
        self.path_index = {}
        self.index_path()
        self.pat_ch_chr = re.compile("[〇一-\u9fff㐀-\u4dbf豈-\ufaff]")

    def bigram_associations(self, subcorp_idx=None, chinese_only=True, 
        sort_by="Gsq", reverse=True, fq_thresh=0):
        distr = self.freq_distr_ngrams(2, subcorp_idx, chinese_only)
        N = sum(distr.values())
        R1 = Counter()
        C1 = Counter()
        for w1w2, fq in distr.items():
            w1, w2 = w1w2[0], w1w2[1]
            R1.update({w1: fq})
            C1.update({w2: fq})
        
        output = []
        for w1w2, o11 in distr.items():
            if o11 < fq_thresh: continue
            w1, w2 = w1w2[0], w1w2[1]
            r1 = R1.get(w1, 0)
            r2 = N - r1
            c1 = C1.get(w2, 0)
            o12 = r1 - o11
            o21 = c1 - o11
            o22 = r2 - o21
            o11_raw = o11
            o11, o12, o21, o22, e11, e12, e21, e22 = \
                additive_smooth(o11_raw, o12, o21, o22, alpha=0)
            stats = { 
                func.__name__: func(o11, o12, o21, o22, e11, e12, e21, e22)\
                    for func in self.association_measures
            }
            stats['RawCount'] = o11_raw
            output.append((w1w2, stats))

        return sorted(output, reverse=reverse, key=lambda x: x[1][sort_by])


    def freq_distr_ngrams(self, n, subcorp_idx=None, chinese_only=True):
        if (not hasattr(self, 'ngrams')) or (n not in self.ngrams):
            self._count_ngrams(n)
        
        if isinstance(subcorp_idx, int):
            distr =  self.ngrams[n].get(subcorp_idx, Counter())
        else:
            for i, key in enumerate(self.ngrams[n]):
                if i == 0: distr = self.ngrams[n][key].copy()
                else: distr.update(self.ngrams[n][key])
        
        if chinese_only: 
            for k in list(distr.keys()):
                if sum(1 for ch in k if self.pat_ch_chr.search(ch)) < n:
                    del distr[k]
        
        return distr


    def _count_ngrams(self, n):
        print(f'Counting {n}-grams...')
        if not hasattr(self, 'ngrams'): self.ngrams = {}
        self.ngrams[n] = {}
        for i in trange(len(self.corpus)):
            self.ngrams[n][i] = Counter()
            for text in self.corpus[i]['text']:
                for sent in text['c']:
                    for ngram in ngrams(sent, n=n):
                        ng = ''.join(ngram)
                        self.ngrams[n][i].update({ng: 1})


    def get_texts(self, pattern, texts_as_str=False, sents_as_str=True):
        texts = {}
        for id in self._list_pattern(pattern):
            text = self.get_text(id, as_str=False)
            if text is None: continue
            if sents_as_str:
                texts[id] = '\n'.join(text)
            else:
                texts[id] = text
        if texts_as_str: 
            return '\n'.join(texts.values())
        return texts
            

    def get_text(self, id, as_str=False):
        idx = self.path_index.get(id, None)
        if idx is None or isinstance(idx, int): 
            return None
        i, j = idx
        text = self.corpus[i]['text'][j].get('c', [])
        if as_str:
            text = '\n'.join(text)
        return text


    def get_meta_by_path(self, id):
        idx = self.path_index.get(id, None)
        if idx is None:
            return {}
        if isinstance(idx, int):
            return self.corpus[idx].get('m', {})
        if isinstance(idx, tuple):
            i, j = idx
            return self.corpus[i]['text'][j].get('m', {})
        return {}


    def list_files(self, pattern, generator=False):
        if generator:
            return self._list_pattern(pattern)
        return list(self._list_pattern(pattern))


    def _list_pattern(self, pattern):
        pattern = re.compile(pattern)
        for k in self.path_index.keys():
            if pattern.search(k):
                yield k


    def index_path(self):
        print("Indexing corpus for text retrival...")
        for i in trange(len(self.corpus)):
            self.path_index[self.corpus[i]['id']] = i
            for j, text in enumerate(self.corpus[i]['text']):
                self.path_index[text['id']] = (i, j)



class IndexedCorpus(TextBasedCorpus):
    """Corpus object for fast concordance search
    """

    def __init__(self, corpus) -> None:
        TextBasedCorpus.__init__(self, corpus)
        self.index = {}
        self.index_corpus()
    

    def get_meta(self, subcorp_idx, text_idx=None, keys:list=None, include_id=True):
        if text_idx is None:
            meta = self.corpus[subcorp_idx]['m']
            if include_id:
                meta['id'] = self.corpus[subcorp_idx]['id']
        else:
            meta = self.corpus[subcorp_idx]['text'][text_idx]['m']
            if include_id:
                meta['id'] = self.corpus[subcorp_idx]['text'][text_idx]['id']
        if keys:
            keys.append('id')
            return { k:meta[k] for k in keys if k in meta }
        return meta


    def index_corpus(self):
        print("Indexing corpus for concordance search...")
        for i in trange(len(self.corpus)):
            for j, text in enumerate(self.corpus[i]['text']):
                for k, sent in enumerate(text['c']):
                    for l, char in enumerate(sent):
                        if char not in self.index:
                            self.index[char] = []
                        self.index[char].append( (i, j, k, l) )




from math import log2, log
from scipy import spatial
# from scipy.stats import fisher_exact
from .fisher_exact import test1t as fisher_exact

def cossim(x, y) -> float:
    """[summary]

    Parameters
    ----------
    x : [type]
        [description]
    y : [type]
        [description]

    Returns
    -------
    float
        [description]
    """
    return 1 - spatial.distance.cosine(x, y)


# Association measures
def MI(O11, O12, O21, O22, E11, E12, E21, E22):
    return log2(O11 / E11)

def Xsq(O11, O12, O21, O22, E11, E12, E21, E22):
    val = (O11 - E11)**2 / E11 + (O12 - E12)**2 / E12 + (O21 - E21)**2 / E21 + (O22 - E22)**2 / E22
    if O11 < E11:
        val = -val
    return val

def Gsq(O11, O12, O21, O22, E11, E12, E21, E22):
    val = 2 * (GsqT(O11, E11) + GsqT(O12, E12) + GsqT(O21, E21) + GsqT(O22, E22))
    if O11 < E11:
        val = -val
    return val

def Dice(O11, O12, O21, O22, E11, E12, E21, E22):
    return 2 * O11 / (O11 + O12 + O11 + O21)

def DeltaP21(O11, O12, O21, O22, E11, E12, E21, E22):
    return O11 / (O11 + O12) - O21 / (O21 + O22)

def DeltaP12(O11, O12, O21, O22, E11, E12, E21, E22):
    return O11 / (O11 + O21) - O12 / (O12 + O22)

def FisherExact(O11, O12, O21, O22, E11, E12, E21, E22):
    pval = fisher_exact(O11, O12, O21, O22)
    if O11 < E11: 
        pval *= -1
    return pval


# Helpers for association measures
def GsqT(O, E):
    if O == 0: return 0
    return O * log(O/E)

def additive_smooth(O11, O12, O21, O22, alpha=0.1):
    O11 += alpha
    O12 += alpha
    O21 += alpha
    O22 += alpha
    N = O11 + O21 + O12 + O22
    R1 = O11 + O12
    C1 = O11 + O21
    R2 = N - R1
    C2 = N - C1
    return O11, O12, O21, O22, R1*C1/N, R1*C2/N, R2*C1/N, R2*C2/N


import torch
import opencc
from pathlib import Path
from transformers import BertTokenizer, BertConfig, BertModel


class AnchiBert:

    def __init__(self, model_path=None):
        if model_path is None:
            model_path = 'AnchiBERT/'
            if not Path(model_path).is_dir():
                _download_bert_model()

        # Bert initialization
        print(f'Loading AnchiBERT model from {model_path} ...')
        config = BertConfig.from_pretrained(model_path)
        self.tokenizer = BertTokenizer.from_pretrained(model_path)
        self.model = BertModel.from_pretrained(model_path,config=config)

        # Traditional to simplified chinese
        self.t2s = opencc.OpenCC('t2s.json')


    def encode_sentence(self, sent:str, idx_from:int=None, idx_to:int=None, is_traditional=True):
        """Encode a raw sentence as a vector

        Parameters
        ----------
        sent : str
            The sentence to encode.
        idx_from : int, optional
            Begining index of a token in the :code:`sent`, 
            by default None.
        idx_to : int, optional
            Ending index of a token in the :code:`sent`, 
            by default None. If :code:`idx_from` or 
            :code:`idx_to` is None, the sentence would be
            encoded by summing all tokens' vectors from the
            last hidden state in the Bert model. Otherwise,
            the subset of tokens in :code:`sent` as specified 
            by :code:`(idx_from, idx_to)` are used.
        is_traditional : bool, optional
            [description], by default True

        Returns
        -------
        numpy.ndarray
            A :code:`(768,)` numpy.ndarray array.
        """
        if is_traditional: sent = self.t2s.convert(sent)
        input = self.tokenizer(sent, return_tensors="pt")
        last_hidden_state = self.model(**input)[0]

        # Sentence vector (sum all tokens together)
        if idx_from is None or idx_to is None:
            return torch.sum(last_hidden_state[0, :, :], dim=0).detach().numpy()
        # Token vector (sum subset of tokens together)
        return torch.sum(last_hidden_state[0, (idx_from+1):(idx_to+2), :], dim=0).detach().numpy()


def _download_bert_model():
    print("Downloading AnchiBert model ...")
    import gdown
    gdown.download('https://drive.google.com/uc?id=1uMlNJzilEhSigIcfjTjPdYOZL9IQfHNK', quiet=False, output="AnchiBERT.zip")
    gdown.extractall("AnchiBERT.zip")



def read_text_as_sentences(fp):
    """Read text file as a sequency of sentences

    Parameters
    ----------
    fp : str
        Path to UTF-8 encoded plain text file.

    Yields
    -------
    str
        A string representing a sentence. A sentence
        corresponds to a line in the file.
    """
    with open(fp, encoding="utf-8") as f:
        for line in f:
            line = line.strip()
            if line == "": continue
            yield line



from math import log2, sqrt
from statistics import mean, pstdev


def Range(d: dict, t:int=1):
    return sum(1 for v in d['v'] if v >= t)


def DP(d: dict):
    V, S, F, n = d['v'], d['s'], d['f'], d['n']
    if F == 0: return 0
    return 0.5 * sum( abs((V[i] / F) - S[i]) for i in range(n) )


def DPnorm(d: dict):
    s = DP(d)
    if s == 0: return 0
    return s / ( 1 - min(d['s']) )


def KLdivergence(d: dict):
    V, S, F, n = d['v'], d['s'], d['f'], d['n']
    stat = 0
    for i in range(n):
        if V[i] == 0: continue
        a = V[i] / F
        stat += a * log2(a / S[i])
    return stat


def JuillandD(d: dict):
    P, n = d['p'], d['n']
    d = mean(P) * sqrt(n - 1)
    if d == 0: return None
    return 1 - ( pstdev(P) / d )


def RosengrenS(d: dict):
    V, S, F, n = d['v'], d['s'], d['f'], d['n']
    if F == 0: return None
    return sum(sqrt(S[i] * V[i]) for i in range(n)) ** 2 / F



# from .postHandler import PostHandler
# from .dirStruct import DirStruct
# from .pipeline import build

# from . import corpusReader
# from . import concordancer
# from . import corpus
# from . import embeddings



import collections


def stringify_obj(x):
    """Stringify complex built-in data structure
    """
    if isinstance(x, dict):
        return '\t'.join(f"{k}: {stringify_obj(v)}" for k, v in x.items())
    if isinstance(x, list) or isinstance(x, tuple):
        return ', '.join(stringify_obj(item) for item in x)
    if isinstance(x, int) or isinstance(x, float):
        return str(x)
    if isinstance(x, str):
        return x
    raise Exception(f'Unexpected data type {type(x)}!')



def flatten(d:dict, parent_key='', sep='.'):
    """Flatten nested dictionary and compress key

    Parameters
    ----------
    d : dict
        A nested dictionary to flatten.
    parent_key : str, optional
        Prefix to prepend to flatten dict keys, by default ''
    sep : str, optional
        Seperator between keys, by default '.'

    Returns
    -------
    dict
        A flatten dictionary
    """
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, collections.MutableMapping):
            items.extend(flatten(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from typing import Callable
from tqdm.auto import trange
from .UtilsGeneral import stringify_obj, flatten
from .UtilsStats import cossim

class ConcordSimil:
    """Similarity analysis of concordance lines
    """

    def __init__(self, concord_lines, bert_model, is_traditional=True, 
        token_idx:tuple=None):
        """Initialize similarity analysis

        Parameters
        ----------
        concord_lines : Sequence
            A sequence of :class:`dcctk.concordancer.ConcordLine`.
        bert_model : AnchiBert
            An intialized :class:`dcctk.embeddings.AnchiBert`. Alternatively,
            one can write a custom class that has a method 
            :code:`encode_sentence`, as in 
            :meth:`dcctk.embeddings.AnchiBert.encode_sentence`.
        is_traditional : bool, optional
            Whether to use :meth:`opencc.Opencc.convert` to translate tradional
            Chinese into simplified Chinese, by default True, which uses the
            :code:`t2s.json` configuration.
        """
        
        self.model = bert_model
        self.concords = concord_lines
        self.concords_len = len(self.concords)
        self.embeddings = np.empty((self.concords_len, 768))
        self._get_embeddings(is_traditional, token_idx)
        self.embeddings_plotting = None
        self.clustering = None

        # Plotting info
        hover_df = []
        for i, x in enumerate(self.concords):
            d = {
                'left': x.data['left'],
                'keyword': x.data['keyword'],
                'right': x.data['right'],
                'timestep': x.get_timestep(),
            }
            for k, v in flatten(x.data['meta'], parent_key='m').items():
                d[k] = stringify_obj(v)
            d['emb_id'] = i
            hover_df.append(d)
        self.hover_df = pd.DataFrame(hover_df)
    

    def __repr__(self) -> str:
        return f"<ConcordSimil  clustered: {'cluster' in self.hover_df}>"


    def hierarchical_clustering(self, threshold=5, criterion='maxclust', standardize_features=True, method='average', metric='cosine', visualize=True):
        """Hierarchical clustering of concordance lines
        """
        from scipy.cluster.hierarchy import fcluster

        if self.clustering is None:
            self.hierarchical_clustering_explore(method=method, metric=metric, standardize_features=standardize_features, dendrogram=False, elbow=False)
        self.hover_df['cluster'] = fcluster(self.clustering, t=threshold, criterion=criterion)

        if visualize:
            self.plot_cluster_results()

        return self.hover_df
    

    def plot_cluster_results(self, interactive=False, labels='cluster', **keywords):
        if labels not in self.hover_df:
            Warning('Clustering not performed yet, please run `ConcordSimil.hierarchical_clustering()` before calling this function.')
            return
        self.plot_embeddings(interactive=interactive, labels=labels, **keywords)


    def hierarchical_clustering_explore(self, method='average', metric='cosine', \
        standardize_features=True, dendrogram=True, elbow=True,
        elbow_metrics="distortion calinski_harabasz silhouette", figsize=(23,7)):
        """Perform hierarchical clustering on the embedding space

        Parameters
        ----------
        standardize_features : bool, optional
            Whether to use zscores for each of the 768 dimension in the 
            Bert model, by default True
        """
        from scipy.stats import zscore
        from scipy.cluster.hierarchy import linkage

        # Scale features
        if standardize_features:
            X = zscore(self.embeddings, axis=0)
        else:
            X = self.embeddings
        
        # Clustering with cosine distance and average linkage method
        self.clustering = linkage(X, method=method, metric=metric)
        # Dendrogram
        if dendrogram:
            self.plot_dendrogram(figsize=figsize)
        
        # Elbow visualizer
        if elbow:
            from sklearn.cluster import AgglomerativeClustering
            model = AgglomerativeClustering(linkage=method, affinity=metric)
            self.plot_elbow(model, X, metrics=elbow_metrics, figsize=figsize)


    def plot_elbow(self, cluster_model, data, metrics='distortion calinski_harabasz silhouette', figsize=(23, 7)):
        from yellowbrick.cluster import KElbowVisualizer
        
        # Plot
        metrics = metrics.strip().split()
        fig, axes = plt.subplots(ncols=len(metrics), figsize=figsize)
        for i, m in enumerate(metrics):
            visualizer = KElbowVisualizer(cluster_model, k=(2,30), metric=m, timings= False, locate_elbow=True, ax=axes[i])
            visualizer.fit(data)      # Fit the data to the visualizer
            visualizer.finalize()  # Finalize and render the figure


    def plot_dendrogram(self, model=None, figsize=(25,10)):
        from scipy.cluster.hierarchy import dendrogram
        
        fig = plt.figure(figsize=figsize)
        if model is None:
            dn = dendrogram(self.clustering)
        else:
            dn = dendrogram(model)
        plt.show()


    def plot_embeddings(self, interactive=True, labels:str=None, **keywords):
        import umap.plot
        if self.embeddings_plotting is None:
            self.embeddings_plotting = self.embeddings_dim_reduce(dim=2, method="umap")
        labels = self.hover_df[labels] if labels is not None else None

        if interactive:
            from bokeh.plotting import show as bokeh_show
            f = umap.plot.interactive(self.embeddings_plotting, 
                            labels=labels, 
                            hover_data=self.hover_df, 
                            **keywords)
            bokeh_show(f)
        else:
            umap.plot.points(self.embeddings_plotting, labels=labels, **keywords)


    def embeddings_dim_reduce(self, dim=2, method="umap"):
        print(f"Reducing the embedding space to {dim} dimensions...")
        if method.lower() == "umap":
            import umap
            emb = umap.UMAP(n_components=dim, metric='cosine').fit(self.embeddings)
        if method.lower() == "pca":
            from sklearn.decomposition import PCA
            emb = PCA(n_components=dim).fit_transform(self.embeddings)
        return emb


    def semantic_sort(self, base_sent:str=None, base_tk:tuple=None, is_traditional=True, simil_func:Callable=cossim):
        """Sort concordance lines based on Bert vector similarities

        Parameters
        ----------
        base_sent : str, optional
            The sentence used to compare with, by default None,
            which takes the first element of :code:`concord_lines`
            as the base sentence.
        base_tk : tuple, optional
            A tuple of length 2, specifying the position 
            :code:`(idx_from, idx_to)` of the keywords in
            :code:`base_sent`, by default None. If None, the
            vector is computed from summing up all token vectors in the sentence.
        is_traditional : bool, optional
            Whether to convert traditional into simplified Chinese
            before feeding the input to AnchiBert, by default True.
            This parameter is passed to 
            :meth:`dcctk.embeddings.AnchiBert.encode_sentence`.
        simil_func : Callable, optional
            Function for computing similarity between two vectors,
            by default :func:`dcctk.UtilsStats.cossim`. The function 
            should take the input vectors as its two arguments and
            returns a float.

        Returns
        -------
        list
            A list of 2-tuples :code:`(<ConcordLine>, <float>)`, 
            with the second elements being the similarity scores
            to :code:`base_sent (base_tk)`.
        """
        if base_sent is None:
            base_sent, base_tk = self.concords[0].get_kwic()
        
        if base_tk is not None:
            compare_base = self.model.encode_sentence(base_sent, *base_tk, is_traditional=is_traditional)
        else:
            compare_base = self.model.encode_sentence(base_sent, is_traditional=is_traditional)

        results = []
        for i in trange(len(self.concords)):
            results.append( (self.concords[i], simil_func(compare_base, self.embeddings[i])) )

        return sorted(results, key=lambda x: x[1], reverse=True)


    def _get_embeddings(self, is_traditional=True, token_idx:tuple=None):
        """Pre-compute embeddings from Bert model

        Parameters
        ----------
        is_traditional : bool, optional
            Whether to convert traditional into simplified Chinese
            before feeding the input to AnchiBert, by default True.
            This parameter is passed to 
            :meth:`dcctk.embeddings.AnchiBert.encode_sentence`.
        token_idx : tuple, optional
            A tuple of length 2, specifying the position 
            :code:`(idx_from, idx_to)` of the keywords in
            :code:`base_sent`, by default None. If None, the
            vector is computed from summing up all token vectors in the sentence.
        """
        print('Computing bert embeddings...')
        for i in trange(self.concords_len):
            sent, tk_idx = self.concords[i].get_kwic()
            if token_idx is not None:
                tk_idx = token_idx
            self.embeddings[i] = self.model.encode_sentence(sent, *tk_idx, is_traditional=is_traditional)



from setuptools import setup
from itertools import chain

with open("README.md", encoding="utf-8") as f:
      long_description = f.read().strip()


with open("VERSION") as f:
      v = int(f.read().strip())

EXTRAS_REQUIRE = {
      'sense': ['transformers', 'torch', 'opencc', 'umap-learn', 'yellowbrick>=1.3'],
      'colab': ['datashader', 'bokeh', 'holoviews', 'scikit-image', 'colorcet'],
}
EXTRAS_REQUIRE['all'] = list(set(chain(*EXTRAS_REQUIRE.values())))

setup(name='dcctk',
      version=f'0.0.{v}',
      description="Diachronic Character-based Corpus toolkit",
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='http://github.com/liao961120/dcctk',
      author='Yongfu Liao',
      author_email='liao961120@github.com',
      license='MIT',
      packages=['dcctk'],
      install_requires=['scikit-learn', 'scipy', 'gdown>=3.10.2', 'pyyaml>=5.1', 'cqls', 'tqdm', 'hanziPhon'],
      dependency_links=[
            'https://github.com/liao961120/CompoTree/tarball/main',
      ],
      extras_require=EXTRAS_REQUIRE,
      classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
      ],
      python_requires='>=3.7',
      # tests_require=['deepdiff'],
      zip_safe=False)



from .UtilsConcord import match_mode

SUBCHAR_ATTRS = set('compo idc radical phon sys'.split())


def is_subchar(tk:dict):
    keys = tk.get('match', {}).keys()
    for k in SUBCHAR_ATTRS:
        if k in keys: return True
    return False    


def has_cql_match_type(query, type_='regex'):
    for tk in query:
        for tp in ['match', 'not_match']:
            for k, values in tk.get(tp, {}).items():
                if k != 'char': continue
                for v in values:
                    if match_mode(v)[-1] == type_:
                        return True
    return False

def has_plain_cql(query):
    for tk in query:
        for k in tk.get('match', {}):
            if k == 'char': return True
        for k in tk.get('not_match', {}):
            if k == 'char': return True
    return False


def all_plain_cql(query):
    for tk in query:
        for tp in ['match', 'not_match']:
            for k in tk.get(tp, {}):
                if k in SUBCHAR_ATTRS: return False
    return True 


from itertools import tee

def ngrams(sequence, n=2):
    """Return the ngrams generated from a sequence of items, as an iterator.

    Parameters
    ----------
    sequence : Sequence
        The source data to be converted into ngrams
    n : int, optional
        The degree of the ngrams, by default 2
    
    Notes
    -----
    Implementation based on nltk.util.ngrams
    https://github.com/nltk/nltk/blob/develop/nltk/util.py#L825
    """
    iterables = tee(sequence, n)
    for i, iterable in enumerate(iterables):
        for _ in range(i):
            next(iterable, None)
    return zip(*iterables)
